
From philb@iag.net Mon Mar 18 18:54:27 1996
Path: news.iag.net!news.math.psu.edu!chi-news.cic.net!news.cic.net!condor.ic.net!gate1.mika.com!news.cencom.net!netnews
From: backbytr@cencom.net
Newsgroups: comp.lang.clipper
Subject: Binary search for fixed length ascii files.
Date: Mon, 18 Mar 96 15:28:41 GMT
Organization: Cen-Com Internet
Lines: 105
Message-ID: <N.031896.092841.74@electra.cencom.net>
NNTP-Posting-Host: electra.cencom.net
X-Newsreader: Quarterdeck Message Center [1.1]

I wrote the following function.  If you can spot any errors, or any 
enhancements that might make the code faster, feel free to do so.

If you want the code, by all means, use it.

***********************************************************


#include "Fileio.ch"

FUNCTION Main( cKey )
   ? FBinarySearch( "q:\dropsrvc.179", cKey, 9, 10, 601 )
RETURN( NIL )

/* BINSRCH.PRG
Function: FBinarySearch()

Program modifications:
  Date   Author
03/17/96 M Jackson
   Original write of function.

Purpose: This function was written to allow searching of a transient file
   downloaded monthly from an IBM 3090.  File space considerations, ( 1.4G ),
   do not allow loading this file into a dbf.  The file is comprised of 2.3
   million fixed length records and is in ascending order.  This routine will
   perform a binary search of that file.

Arguments:
   cFile      - Path and filename of file.
   cKey       - The key to be searched for.
   nKeyLen    - The length of the key in the record.
   nKeyOffset - Starting position of the key in the record.
   nRecLen    - The record length.

Returns: 1. A record matching the passed key.
         2. An empty string indicating no record found for the passed key.
         3. A string beginning with the characters 'ERROR--'.

Known problems:
   1. I don't return the value of FERROR().
   2. If the file contains duplicate records, you will get one of those
      records, but it is impossible to tell which one.
Possible additions:
   1. Ascending/descending flag.
   2. Pass an array of key information.
   3. Return an array of records matching the key

*/
FUNCTION FBinarySearch( cFile, cKey, nKeyLen, nKeyOffset, nRecLen )
   LOCAL cResult := ""                           // Empty means we found nothing.
   LOCAL lDone   := .F.                          // Loop controller
   LOCAL nHandle                                 // File handle for low-level I-O
   LOCAL nSize   := 0                            // File size in bytes.
   LOCAL nMaxRecords := 0                        // Total number of records in file
   LOCAL nLowerBound := 0                        // Lower range of next search
   LOCAL nUpperBound := 0                        // Upper range of next search
   LOCAL nCurrentRecord := 0                     // Current record number of search
   LOCAL cRecord := SPACE( nRecLen )             // Our record for FREAD()

   nHandle := FOPEN( cFile, FO_READ + FO_SHARED) // Open the file
   IF FERROR() != 0                              // Check for errors
      cResult := "ERROR-- Cannot open file!"
   ELSE
      nSize := FSEEK( nHandle, 0, FS_END )       // Getting the size of the file
      IF nSize < nRecLen                         // Making sure we have at least one record
         cResult := "ERROR-- Bad file size!"
      ELSE
         nMaxRecords := nSize / nRecLen          // Getting the number of records
         nLowerBound := 1                        // Setting lower search area to beginning of file
         nUpperBound := nMaxrecords              // Setting upper search area to end-of-file
         lDone := .F.                            // Set Loop controller to false
         DO WHILE ! lDone
            nCurrentRecord := INT( ( nLowerBound + nUpperBound ) / 2 ) // Find halfway point
            FSEEK( nHandle, nCurrentRecord * nRecLen - nRecLen, FS_SET ) // Go there
            IF FREAD( nHandle, @cRecord, nRecLen ) <> nRecLen // Read the record
               cResult := "ERROR-- Bad read of file!" // Not a good read
               lDone := .T.                      // Exiting loop
            ELSE
               IF cKey == SUBSTR( cRecord, nKeyOffset, nKeyLen ) // Is this it?
                  cResult := cRecord             // Send the record along
                  lDone := .T.                   // Exiting loop
               ELSE
                  IF cKey > SUBSTR( cRecord, nKeyOffset, nKeyLen ) // Which direction
                     nLowerBound := nCurrentRecord + 1 // Search higher
                  ELSE
                     nUpperBound := nCurrentRecord - 1 // Search lower
                  ENDIF
                  IF nLowerBound > nUpperBound   // We didn't find it
                     cResult := cKey + " -- NOT FOUND"
                     lDone := .T.                // Exiting loop
                  ENDIF
               ENDIF
            ENDIF
         ENDDO
      ENDIF
   ENDIF
   FCLOSE( nHandle )                             // Close the file
RETURN( cResult )

